; ============================
;  Fellow Amiga Emulator      
;  Virtual Memory
;  (C) 1994-1998 Petter Schau
; ============================



        .486p
DGROUP group _DATA,_BSS
_TEXT  segment para public use32 'CODE'
       assume cs:_TEXT, ds:DGROUP, ss:DGROUP

public memorystart_
memorystart_:                   ; Used to check code alignment
startlbl:

include defs.inc
include func\68000.inc
include func\inout.inc
include func\sound.inc
include func\draw.inc
include func\graphem.inc
include func\copper.inc
include func\blit.inc
include mac\memory.mac
		
#ifdef DEBUGBUILD
include eventid.inc
#endif
	
public oddread
public oddwrite

; Export to c-routines
; --------------------


public readregisterbyte_
public readregisterword_
public readregisterlong_
public writeregisterbyte_
public writeregisterword_
public writeregisterlong_
public rintenar_
public rintreqr_
public rdefault_
public wcopcon_
public wcop1lch_
public wcop1lcl_
public wcop2lch_
public wcop2lcl_
public wcopjmp1_
public wcopjmp2_
public wdmacon_
public wintena_
public wintreq_
public wadcon_
public wbplcon0_
public wbplcon1_
public wbplcon2_
public wbpl1mod_
public wbpl2mod_
public wspr0pth_
public wspr0ptl_
public wspr1pth_
public wspr1ptl_
public wspr2pth_
public wspr2ptl_
public wspr3pth_
public wspr3ptl_
public wspr4pth_
public wspr4ptl_
public wspr5pth_
public wspr5ptl_
public wspr6pth_
public wspr6ptl_
public wspr7pth_
public wspr7ptl_

public wdefault_


;
; Fetches a byte from memory, returns byte in dl.  Takes ecx as <ea>
;


		ALIGN32

public fetb_
fetb_:		push	ecx
		push	edx
		mov	ecx, eax
		call	fetb
		mov	eax, edx
		pop	edx
		pop	ecx
		ret


		ALIGN32

public fetb
fetb:		mov	edx, ecx
		shr	edx, 16
		jmp	dword ptr [_memory_bank_readbyte + 4*edx]


;
; Fetches a word from memory, returns word in dx.  Takes ecx as <ea>
;


		ALIGN32

public fetw_
fetw_:		push	ecx
		push	edx
		mov	ecx, eax
		call	fetw
		mov	eax, edx
		pop	edx
		pop	ecx
		ret


		ALIGN32

public fetw
fetw:		TEST_ODD_ACCESS	oddread
		mov	edx, ecx
		shr	edx, 16
		jmp	dword ptr [_memory_bank_readword + 4*edx]


;
; Fetches a long from memory, returns long in edx.  Takes ecx as <ea>
;


		ALIGN32

public fetl_
fetl_:		push	ecx
		push	edx
		mov	ecx, eax
		call	fetl
		mov	eax, edx
		pop	edx
		pop	ecx
		ret


		ALIGN32

public fetl
fetl:		TEST_ODD_ACCESS	oddread
		mov	edx, ecx
		shr	edx, 16
		jmp	dword ptr [_memory_bank_readlong + 4*edx]

;
; Writes a byte to memory, takes data in dl, ea in ecx
;


		ALIGN32

public wrib_
wrib_:		push	ecx
		push	edx
		mov	ecx, edx
		mov	edx, eax
		call	wrib
		pop	edx
		pop	ecx
		ret


		ALIGN32

public wrib
wrib:		mov	dword ptr [_wriorgadr], ecx
		shr	ecx, 16
		jmp	dword ptr [_memory_bank_writebyte + 4*ecx]


;
; Writes a word to memory, takes data in dx, ea in ecx
;


		ALIGN32

public wriw_
wriw_:		push	ecx
		push	edx
		mov	ecx, edx
		mov	edx, eax
		call	wriw
		pop	edx
		pop	ecx
		ret


		ALIGN32

public wriw
wriw:		TEST_ODD_ACCESS	oddwrite
		mov	dword ptr [_wriorgadr], ecx
		shr	ecx, 16
		jmp	dword ptr [_memory_bank_writeword + 4*ecx]


;
; Writes a long to memory, takes data in edx, ea in ecx
;


		ALIGN32

public wril_
wril_:		push	ecx
		push	edx
		mov	ecx, edx
		mov	edx, eax
		call	wril
		pop	edx
		pop	ecx
		ret


		ALIGN32

public wril
wril:		TEST_ODD_ACCESS	oddwrite
		mov	dword ptr [_wriorgadr], ecx
		shr	ecx, 16
		jmp	dword ptr [_memory_bank_writelong + 4*ecx]


;
; These raises exception 3 when reading an odd address
;


		ALIGN32

oddread:	mov	dword ptr [_memory_fault_read] , 1
		mov	dword ptr [_memory_fault_address], ecx
		mov	ebx, 0ch
		mov	eax, 44
		jmp	prepare_exception_


		ALIGN32

oddwrite:	mov	dword ptr [_memory_fault_read] , 0
		mov	ecx, dword ptr [_wriorgadr]
		mov	dword ptr [_memory_fault_address], ecx
		mov	ebx, 0ch
		mov	eax, 44
		jmp	prepare_exception_


; ================================================= 
; These functions are the bank functions            
; called from the fetX wriX functions               
; ================================================= 

; --------- 
; Chip bank 
; --------- 


		ALIGN32

public readchipbyte_
readchipbyte_:  and	ecx, 0ffffffh
		movzx	edx, byte ptr [_cmem + ecx]
		ret


		ALIGN32

public readchipword_
readchipword_:	and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [_cmem + ecx]
		mov	dl, byte ptr [_cmem + 1 + ecx]
		ret


		ALIGN32

public readchiplong_
readchiplong_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_cmem + ecx]
		bswap	edx
		ret

	
		ALIGN32

public writechipbyte_
writechipbyte_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	byte ptr [_cmem + ecx], dl
		ret


		ALIGN32

public writechipword_
writechipword_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	byte ptr [_cmem + ecx], dh
		mov	byte ptr [_cmem + 1 + ecx], dl
		ret


		ALIGN32

public writechiplong_
writechiplong_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		bswap	edx
		mov	dword ptr [_cmem + ecx], edx
		ret

; ---------
; Fast bank 
; ---------


		ALIGN32

public readfastbyte_
readfastbyte_:	mov	edx, dword ptr [_fmem]
		and	ecx, 0ffffffh
		movzx	edx, byte ptr [ecx + edx - 0200000h]
		ret


		ALIGN32

public readfastword_
readfastword_:	push	ebx
		mov	ebx, dword ptr [_fmem]
		and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [ecx + ebx - 0200000h]
		mov	dl, byte ptr [ecx + ebx - 01fffffh]
		pop	ebx
		ret


		ALIGN32

public readfastlong_
readfastlong_:	mov	edx, dword ptr [_fmem]
		and	ecx, 0ffffffh
		mov	edx, dword ptr [ecx + edx - 0200000h]
		bswap	edx
		ret


		ALIGN32

public writefastbyte_
writefastbyte_:	push	ebx
		mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	ebx, dword ptr [_fmem]
		mov	byte ptr [ecx + ebx - 0200000h], dl
		pop	ebx
		ret


		ALIGN32

public writefastword_
writefastword_:	push	ebx
		mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	ebx, dword ptr [_fmem]
		mov	byte ptr [ecx + ebx - 0200000h], dh
		mov	byte ptr [ecx + ebx - 01fffffh], dl
		pop	ebx
		ret


		ALIGN32

public writefastlong_
writefastlong_:	push	ebx
		mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	ebx, dword ptr [_fmem]
		bswap	edx
		mov	dword ptr [ecx + ebx - 0200000h], edx
		pop	ebx
		ret

; --------- 
; Bogo bank 
; ---------


		ALIGN32

public readbogobyte_
readbogobyte_:	and	ecx, 0ffffffh
		movzx	edx, byte ptr [_bmem - 0c00000h + ecx]
		ret


		ALIGN32

public readbogoword_
readbogoword_:	and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [_bmem - 0c00000h + ecx]
		mov	dl, byte ptr [_bmem - 0c00000h + 1 + ecx]
		ret


		ALIGN32

public readbogolong_
readbogolong_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_bmem - 0c00000h + ecx]
		bswap	edx
		ret


		ALIGN32

public writebogobyte_
writebogobyte_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	byte ptr [_bmem - 0c00000h + ecx], dl
		ret


		ALIGN32

public writebogoword_
writebogoword_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		mov	byte ptr [_bmem - 0c00000h + ecx], dh
		mov	byte ptr [_bmem - 0c00000h + 1 + ecx], dl
		ret


		ALIGN32

public writebogolong_
writebogolong_:	mov	ecx, dword ptr [_wriorgadr]
		and	ecx, 0ffffffh
		bswap	edx
		mov	dword ptr [_bmem - 0c00000h + ecx], edx
		ret


; ------------
; AF Kick bank 
; ------------


		ALIGN32

public readkickbyteaf_
readkickbyteaf_:and	ecx, 0ffffffh
		movzx	edx, byte ptr [_kmem - 0f80000h + ecx]
		ret


		ALIGN32

public readkickwordaf_
readkickwordaf_:and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [_kmem - 0f80000h + ecx]
		mov	dl, byte ptr [_kmem - 0f80000h + 1 + ecx]
		ret


		ALIGN32

public readkicklongaf_
readkicklongaf_:and	ecx, 0ffffffh
		mov	edx, dword ptr [_kmem - 0f80000h + ecx]
		bswap	edx
		ret


; ---------
; Kick bank 
; ---------


		ALIGN32

public readkickbyte_
readkickbyte_:	and	ecx, 0ffffffh
		movzx	edx, byte ptr [_kmem - 0f80000h + ecx]
		ret


		ALIGN32

public readkickword_
readkickword_:	and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [_kmem - 0f80000h + ecx]
		mov	dl, byte ptr [_kmem - 0f80000h + 1 + ecx]
		ret


		ALIGN32

public readkicklong_
readkicklong_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_kmem - 0f80000h + ecx]
		bswap	edx
		ret


		ALIGN32

public writekickbyte_
public writekickword_
public writekicklong_
writekickbyte_:
writekickword_:
writekicklong_:	mov	ecx, dword ptr [_wriorgadr]
		ret

		
; -------------
; Low Kick bank 
; -------------


		ALIGN32

public readlkickbyte_
readlkickbyte_:	and	ecx, 0ffffffh
		movzx	edx, byte ptr [_kmem - 0f00000h + ecx]
		ret


		ALIGN32

public readlkickword_
readlkickword_:	and	ecx, 0ffffffh
		xor	edx, edx
		mov	dh, byte ptr [_kmem - 0f00000h + ecx]
		mov	dl, byte ptr [_kmem - 0f00000h + 1 + ecx]
		ret


		ALIGN32

public readlkicklong_
readlkicklong_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_kmem - 0f00000h + ecx]
		bswap	edx
		ret


		ALIGN32


	
public readk2byte_
readk2byte_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_k2mem]
		movzx	edx, byte ptr [ecx + edx - 0f80000h]
		ret


		ALIGN32

public readk2word_
readk2word_:	and	ecx, 0ffffffh
		add	ecx, dword ptr [_k2mem]
		xor	edx, edx
		mov	dh, byte ptr [ecx - 0f80000h]
		mov	dl, byte ptr [1 - 0f80000h + ecx]
		sub	ecx, dword ptr [_k2mem]
		ret


		ALIGN32

public readk2long_
readk2long_:	and	ecx, 0ffffffh
		mov	edx, dword ptr [_k2mem]
		mov	edx, dword ptr [ecx + edx - 0f80000h]
		bswap	edx
		ret


	
; ----------------
; Unavailable bank
; ----------------


		ALIGN32

public readNAbyte_
readNAbyte_:	not	dword ptr [_mysteryv]
		mov	edx, dword ptr [_mysteryv]
		ret


		ALIGN32

public readNAword_
readNAword_:	mov	edx, 06100h
		ret
        

		ALIGN32

public readNAlong_
readNAlong_:	mov	edx, 061006100h
		ret


		ALIGN32
	
public writeNAbyte_
public writeNAword_
public writeNAlong_
writeNAbyte_:
writeNAword_:
writeNAlong_:	mov	ecx, dword ptr [_wriorgadr]
		ret

; ----------------------
; Mystery (SCSI..?) bank
; ----------------------


		ALIGN32

public readMysterybyte_
public readMysteryword_
public readMysterylong_
readMysterybyte_:
readMysteryword_:
readMysterylong_:
		not	dword ptr [_mysteryv]
		mov	edx, dword ptr [_mysteryv]
		ret


		ALIGN32
	
public writeMysterybyte_
public writeMysteryword_
public writeMysterylong_
writeMysterybyte_:
writeMysteryword_:
writeMysterylong_:
		mov	ecx, dword ptr [_wriorgadr]
		ret


; ------------------- 
; Chip register bank  
; ------------------- 

        ALIGN32
readregisterbyte_:
        push    ecx
        and     ecx,01feh
        call    dword ptr _ioread[ecx*2]
        pop     ecx
        test    ecx,1
        jnz     rrb1
        shr     edx,8
rrb1:   and     edx,0ffh
        ret

        ALIGN32
readregisterword_:
        push    ecx
        and     ecx,01feh
        call    dword ptr _ioread[ecx*2]
        pop     ecx
        ret

        ALIGN32
readregisterlong_:
        push    ecx
        and     ecx,01feh
        push    ecx
        call    dword ptr _ioread[ecx*2]
        pop     ecx
        push    edx
        call    dword ptr _ioread+4[ecx*2]
        pop     ecx
        shl     ecx,16
        or      edx,ecx
        pop     ecx
        ret


; Byte write to register, well, simply clear the other
; byte and do a word write.

        ALIGN32
writeregisterbyte_:
        mov     ecx,dword ptr _wriorgadr
        push    ecx
        test    ecx,1
        jnz     wrbodd
        mov     dh,dl

        and     ecx,01feh
        and     edx,0ff00h
        call    dword ptr _iowrite[ecx*2]
        pop     ecx
        ret
wrbodd:
        and     ecx,01feh
        and     edx,0ffh
        call    dword ptr _iowrite[ecx*2]
        pop     ecx
        ret

	
        ALIGN32
writeregisterword_:
        mov     ecx,dword ptr _wriorgadr
        push    ecx
        and     ecx,01feh
        and     edx,0ffffh              ; No extra bits
        call     dword ptr _iowrite[ecx*2]
        pop     ecx
        ret

        ALIGN32
writeregisterlong_:
        mov     ecx,dword ptr _wriorgadr
        push    ecx
        push    edx
        and     ecx,01feh
        push    ecx
        add     ecx,2
        and     ecx,01feh    ; Unreal writes long to $1fe -> jump beyond array 
        and     edx,0ffffh   ; No extra bits here
        call    dword ptr _iowrite[ecx*2]
        pop     ecx
        pop     edx
        shr     edx,16          ; No extra bits! Good!
        call    dword ptr _iowrite[ecx*2]
        pop     ecx
        ret

; ----------------------------
; Chip read register functions  
; ---------------------------- 

; SERDATR
; $dff018

public rserdatr_
rserdatr_:      xor     edx,edx
                ret

; INTENAR
; $dff01c
                ALIGN32

rintenar_:
        mov     edx,dword ptr _intenar
        ret

; INTREQR
; $dff01e
                ALIGN32

rintreqr_:
        mov     edx,dword ptr [_intreq]
        ret

; To simulate bus-noise, return 0 and -1 every second time.
; Why? Bugged demos test write-only register for various bit-values
; and to break out of loops, both 0 and 1 values must be returned.

                ALIGN32

rdefault_:
        inc     dword ptr [_busnoisecounter]
        mov     edx,dword ptr [_busnoisecounter]
        and     edx,1
        mov     edx,dword ptr [_busnoise+4*edx]
        ret

; -----------------------------
; Chip write register functions  
; ----------------------------- 

;========
; COPCON
;========
; $dff02e

                ALIGN32

wcopcon_:
        mov     dword ptr _copcon,edx
        ret


;=======
; COP1LC
;=======

; $dff080 - WORD
                ALIGN32

wcop1lch_:
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _cop1lc+2,dx
        ret

; $dff082 - WORD
                ALIGN32

wcop1lcl_:
        and     edx,0fffffffeh              ; No odd adresses
        mov     word ptr _cop1lc,dx
        ret

;=======
; COP2LC
;=======

; $dff084 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff084 - WORD
                ALIGN32

wcop2lch_:
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _cop2lc+2,dx
        ret

; $dff086 - WORD
                ALIGN32

wcop2lcl_:
        and     edx,0fffffffeh              ; No odd adresses
        mov     word ptr _cop2lc,dx
        ret

;========
; COPJMP1
;========
; $dff088 - word
                ALIGN32

wcopjmp1_:
        jmp    copper_load1_


public rcopjmp1_
rcopjmp1_:      call   copper_load1_
                xor    edx,edx
                ret

;========
; COPJMP2
;========
; $dff08A - word
                ALIGN32

wcopjmp2_:
        jmp    copper_load2_
        

public rcopjmp2_
rcopjmp2_:      call   copper_load2_
                xor    edx,edx
                ret



; Calculate all variables connected to the screen emulation
; First calculate the ddf variables DDFstartpos and DDFnumberofwords
;   Check if stop is bigger than start
;   Check if strt and stop is mod 8 aligned
;   If not OK then set numberofwords and startpos to 0 and DIWvisible vars to 128


                ALIGN32

public calculate_window_
calculate_window_:      test    byte ptr [_bplcon0 + 1],080h
                        jnz     calculate_window_hires
                        mov     ecx,dword ptr [_ddfstop]
                        mov     edx,dword ptr [_ddfstrt]
                        cmp     edx,ecx         ; ddfstrt bigger than ddfstop?
                        ja      cwdiwerr
                        and     ecx,07h         ; Must be aligned by mod 8
                        and     edx,07h
                        push    ebp
                        xor     ebp,ebp
                        cmp     ecx,edx         ; start aligned with stop?
                        je      cwdiwloddfaligned
                        inc     ebp

; everything seems OK with DDF, calculate DFFstartpos, DDFnumberofwords

cwdiwloddfaligned:      mov     ecx,dword ptr [_ddfstop]
                        sub     ecx,dword ptr [_ddfstrt]
                        shr     ecx,3           ; length in words
                        inc     ecx
                        add     ecx,ebp
                        pop     ebp
                        mov     dword ptr [_DDFnumberofwords],ecx
                        mov     ecx,dword ptr [_ddfstrt]
                        shl     ecx,1
                        add     ecx,11h
                        mov     dword ptr [_DDFstartpos],ecx

; Now DDF variables are done
; Now check the DIW values
; DIW values are only used by the drawing routines, include
; clip modification here as well.

                        mov     ecx,dword ptr [_diwxleft]
                        mov     edx,dword ptr [_diwxright]
                        cmp     ecx,edx         ;diwxleft less than diwxright?
                        jae     cwdiwerr
                        cmp     ecx,dword ptr [_DDFstartpos]
                        jb      cwdiwless
                        cmp     ecx,dword ptr [_clipleftx]
                        ja      cwdiwnoclip
                        mov     ecx,dword ptr [_clipleftx]
cwdiwnoclip:            mov     dword ptr [_DIWfirstvisiblepos],ecx ; DDFstartpos is less than diwxleft
                        jmp     cwdiwlastpos
cwdiwless:              mov     ecx,dword ptr [_DDFstartpos]
                        cmp     ecx,dword ptr [_clipleftx]
                        ja      cwdiwnoclip2
                        mov     ecx,dword ptr [_clipleftx]
cwdiwnoclip2:           mov     dword ptr [_DIWfirstvisiblepos],ecx ; DDFstartpos is higher than diwxleft
cwdiwlastpos:           mov     ecx,dword ptr [_DDFnumberofwords]   ; Calculate last pos in DDF data
                        mov     edx,dword ptr [_DDFstartpos]
                        shl     ecx,4                             ; numofwords*16
                        add     ecx,edx
                        mov     edx,dword ptr [_oddscroll]
                        cmp     edx,dword ptr [_evenscroll]
                        ja      cwaddodd
                        add     ecx,dword ptr [_evenscroll]
                        jmp     cwdiwlastpos2
cwaddodd:               add     ecx,dword ptr [_oddscroll]
cwdiwlastpos2:
; Now we have ecx - the last position in line(first invisible)

                        mov     edx, dword ptr [_diwxright]
                        cmp     ecx,edx
                        jb      cwdiwxx
                        cmp     edx,dword ptr [_cliprightx]
                        jb      cwdiwnoclip3
                        mov     edx,dword ptr [_cliprightx]
cwdiwnoclip3:           mov     dword ptr [_DIWlastvisiblepos],edx ; lastpos is higher than diwxright
                        jmp     cwdiwyy
cwdiwxx:                cmp     ecx,dword ptr [_cliprightx]
                        jb      cwdiwnoclip4
                        mov     ecx,dword ptr [_cliprightx]
cwdiwnoclip4:           mov     dword ptr [_DIWlastvisiblepos],ecx ; lastpos is lower than diwxright
cwdiwyy:

; Now all variables have been initialized

                        ret


                ALIGN32

calculate_window_hires:
        mov     ecx,dword ptr _ddfstop
	mov	edx,dword ptr _ddfstrt
        cmp     edx,ecx                    ; ddfstrt bigger than ddfstop?
        ja      cwdiwerrh
        and     ecx,07h         ; Must be aligned by mod 8 in hires
        and     edx,07h

cwdiwddfaligned:
        
; everything seems OK with DDF, calculate DFFstartpos, DDFnumberofwords
; DDFnumberofwords is  stop-start/4+2
        mov     ecx,dword ptr _ddfstop
        sub     ecx,dword ptr _ddfstrt
        add     ecx,15
        shr     ecx,2             ; length in words
        and     ecx,0fffffffeh
        mov     dword ptr _DDFnumberofwords,ecx

; DDFstartpos = start*4+18 in hires
        mov     ecx,dword ptr _ddfstrt
        shl     ecx,2
        add     ecx,18
        mov     dword ptr _DDFstartpos,ecx

; Now DDF variables are done
; Now check the DIW values

	mov	ecx,dword ptr _diwxleft
        shl     ecx,1        
	mov	edx,dword ptr _diwxright
        shl     edx,1
        cmp	ecx,edx				; diwxleft less than diwxright?
        jae     cwdiwerrh
	cmp	ecx,dword ptr _DDFstartpos
        jb      cwdiwlessh
        mov     edx,dword ptr [_clipleftx]
        shl     edx,1
        cmp     ecx,edx
        ja      cwdiwnoclip5
        mov     ecx,edx
cwdiwnoclip5:
        mov     dword ptr _DIWfirstvisiblepos,ecx ; DDFstartpos is less than diwxleft
        jmp     cwdiwlastposh
cwdiwlessh:
	mov	ecx,dword ptr _DDFstartpos
        mov     edx,dword ptr [_clipleftx]
        shl     edx,1
        cmp     ecx,edx
        ja      cwdiwnoclip6
        mov     ecx,edx
cwdiwnoclip6:
	mov	dword ptr _DIWfirstvisiblepos,ecx ; DDFstartpos is higher than diwxleft
cwdiwlastposh:
 	mov	ecx,dword ptr _DDFnumberofwords   ; Calculate last pos in DDF data
	mov	edx,dword ptr _DDFstartpos
	shl	ecx,4				  ; numofwords*16
	add	ecx,edx

        mov     edx,dword ptr _oddhiscroll
        cmp     edx,dword ptr _evenhiscroll
        ja      cwaddoddh
        add     ecx,dword ptr _evenhiscroll
        jmp     cwdiwlastpos2h
cwaddoddh:
        add     ecx,dword ptr _oddhiscroll
cwdiwlastpos2h:
; Now we have ecx - the last position in line(first invisible)

	mov	edx, dword ptr _diwxright
        shl     edx,1
	cmp	ecx,edx
        jb      cwdiwxxh
        mov     ecx,dword ptr [_cliprightx]
        shl     ecx,1
        cmp     edx,ecx
        jb      cwdiwnoclip7
        mov     edx,ecx
cwdiwnoclip7:
	
	mov	dword ptr _DIWlastvisiblepos,edx ; lastpos is higher than diwxright
        jmp     cwdiwyyh
cwdiwxxh:
        mov     edx,dword ptr [_cliprightx]
        shl     edx,1
        cmp     ecx,edx
        jb      cwdiwnoclip8
        mov     ecx,edx
cwdiwnoclip8:
	mov	dword ptr _DIWlastvisiblepos,ecx ; lastpos is lower than diwxright
cwdiwyyh:

; Now all variables have been initialized

        ret

; Error with DDF variables, set all variables to off values
                ALIGN32

cwdiwerr:
        xor     ecx,ecx
        mov     dword ptr _DDFstartpos,ecx
        mov     dword ptr _DDFnumberofwords,ecx
	mov	dword ptr _DIWfirstvisiblepos,256
	mov	dword ptr _DIWlastvisiblepos,256
        ret
; Error with DDF variables, set all variables to off values
cwdiwerrh:
        xor     ecx,ecx
        mov     dword ptr _DDFstartpos,ecx
        mov     dword ptr _DDFnumberofwords,ecx
	mov	dword ptr _DIWfirstvisiblepos,256
	mov	dword ptr _DIWlastvisiblepos,256
        ret



;========
; DMACON
;========
; $dff096  - Read from $dff002
; dmacon - zero if master dma bit is off
; dmaconr - is always correct.
                ALIGN32

wdmacon_:
        mov     ecx,dword ptr _dmaconr
        test    edx,08000h
	jz      wdma1
        and     edx,7ffh                ; Sett bits - Readonly bits
        or      ecx,edx                 ; ecx is new dmaconr
        mov     dword ptr _dmaconr,ecx
        test    ecx,0200h
	jnz     wdmanorm
	xor     ecx,ecx
wdmanorm:
        mov     edx,dword ptr _dmacon   ; get old dmacon (master dma adjusted)
        mov     dword ptr _dmacon,ecx   ; Store new dmacon --------"---------


; Do audio stuff
        test    ecx,1                   ; Channel 0 turned on?
        jz      snden0
        test    edx,1                   ; Already on?
        jnz     snden0
        push    edx
        push    ecx
        mov     edx,0
	push	eax
        call    audiostate0_
	pop	eax
        pop     ecx
        pop     edx
snden0:
        test    ecx,2
        jz      snden1
        test    edx,2
        jnz     snden1
        push    edx
        push    ecx
#ifndef SOUND_C
        mov     edx,4
#else
	mov	edx,1
#endif
	push	eax
        call    audiostate0_
	pop	eax
        pop     ecx
        pop     edx
snden1:
        test    ecx,4
        jz      snden2
        test    edx,4
        jnz     snden2
        push    edx
        push    ecx
#ifndef SOUND_C
        mov     edx,8
#else
	mov	edx,2
#endif
	push	eax
        call    audiostate0_
	pop	eax
        pop     ecx
        pop     edx
snden2:
        test    ecx,8
        jz      snden3
        test    edx,8
        jnz     snden3
        push    edx
        push    ecx
#ifndef SOUND_C
        mov     edx,12
#else
	mov	edx,3
#endif
	push	eax
        call    audiostate0_
	pop	eax
        pop     ecx
        pop     edx
snden3:
        test    dword ptr _blitterdmawaiting,1
        jz      wdmae1
        test    ecx,0040h
        jz      wdmae1
        call    blitter_copy_
wdmae1:
        call    copper_updatedma_
        ret

wdma1:  and     edx,7ffh                ; Vekk med Readonly bits
        not     edx
        and     ecx,edx                   ; Slett bits
        mov     dword ptr _dmaconr,ecx
        test    ecx,0200h
        jnz     wdmanormoff
	xor     ecx,ecx
wdmanormoff:
        mov     edx,dword ptr _dmacon  ; get old dmacon
        mov     dword ptr _dmacon,ecx  ; store new dmacon

; IF blitter dma is turned off in the middle of a blit
; finish the blit

        test    ecx,040h
        jnz     wdmabrag
        test    dword ptr _blitterstatus,0ffh
        jz      wdmabrag
        pushad
        call    finish_blit_
        popad
wdmabrag:
; Do audio stuff when dma turns off
; Simply set state 0, (state not set if state = 3??)

        test    ecx,1                   ; Channel 0 turned off?
        jnz     sndoff0
        test    edx,1                   ; Already off?
        jz      sndoff0
        push    ecx
        mov     ecx,0
        call    audio_kill_channel_
        pop     ecx
sndoff0:
        test    ecx,2
        jnz     sndoff1
        test    edx,2
        jz      sndoff1
        push    ecx
        mov     ecx,1
        call    audio_kill_channel_
        pop     ecx
sndoff1:
        test    ecx,4
        jnz     sndoff2
        test    edx,4
        jz      sndoff2
        push    ecx
        mov     ecx,2
        call    audio_kill_channel_
        pop     ecx
sndoff2:
        test    ecx,8
        jnz     sndoff3
        test    edx,8
        jz      sndoff3
        push    ecx
        mov     ecx,3
        call    audio_kill_channel_
        pop     ecx
sndoff3:


        test    dword ptr _blitterdmawaiting,1
        jz      wdmae2
        test    ecx,0040h
        jz      wdmae2
        pushad
        call    blitter_copy_
        popad
wdmae2:
        call    copper_updatedma_
        ret

;========
; INTENA
;========
; $dff09a  - Leses fra $dff01c
; If master bit is off, then INTENA is 0, else INTENA = INTENAR
; The master bit can not be read, the memory test in the kickstart
; depends on this.
                ALIGN32

wintena_:
        mov     ecx,dword ptr _intenar     ; Les inn status
        test    edx,08000h
	jz      wine1
        and     edx,7fffh                ; Sett bits
        or      ecx,edx
        mov     dword ptr _intenar,ecx
        test    ecx,00004000h
        jnz     wintenanorm
	xor     ecx,ecx
        mov     dword ptr _intena,ecx
        ret
wintenanorm:
        mov     dword ptr _intena,ecx
        call    raise_interrupt_
	ret
wine1:  and     edx,7fffh                ; Slett bits
        not     edx
        and     ecx,edx                   ; Slett bits
        mov     dword ptr _intenar,ecx
        test    ecx,00004000h
        jnz     wintenanorm
	xor     ecx,ecx
        mov     dword ptr _intena,ecx
	ret

;========
; INTREQ
;========

; $dff09c  - Leses fra $dff01e
; Skrives direkte til $dff01e, og raise_interrupt kalles ved setting av bits.
                ALIGN32

wintreq_:
        mov     ecx,dword ptr _intreq    ; Les inn status
        test    edx,08000h
        jz      winr1
        and     edx,7fffh                ; Sett bits
        or      ecx,edx
        mov     dword ptr _intreq,ecx
        call    raise_interrupt_
        ret
winr1:  and     edx,7fffh                ; Slett bits
        not     edx
        and     ecx,edx                   ; Slett bits
        mov     dword ptr _intreq,ecx
	ret
;========
; ADCON
;========

; $dff09e  - Leses fra $dff010
                ALIGN32

wadcon_:
        mov     ecx,dword ptr _adcon    ; Les inn status
        test    edx,08000h
        jz      adcr1
        and     edx,7fffh                ; Sett bits
        or      ecx,edx
        mov     dword ptr _adcon,ecx
        ret
adcr1:  and     edx,7fffh                ; Slett bits
        not     edx
        and     ecx,edx                   ; Slett bits
        mov     dword ptr _adcon,ecx
        ret

;========
; BPLCON0
;========
; $dff100 - WORD
                ALIGN32

wbplcon0_:
        push    ecx
	
#ifdef BPLCON0_KLUDGE	
	test	byte ptr [_bplcon0 + 1], 0xf0h
	jnz	wbplcon0nokludge
	test	dh,0xf0
	jz	wbplcon0nokludge
	test	byte ptr [_playfieldon],1
	jz	wbplcon0nokludge
	push	ecx
	mov	ecx,dword ptr [_xpos]
	sub	ecx,16
	jle	wbplcon0nokludge2
	cmp	ecx,dword ptr [_ddfstrt]
	jbe	wbplcon0nokludge2
	add	ecx,8
	cmp	ecx,dword ptr [_ddfstop]
	jae	wbplcon0nokludge2

	test	dh,0x80
	jnz	wbplcon0hireskludge

	and	ecx,0f8h
	sub	ecx,dword ptr [_ddfstrt]
	shr	ecx,2

	push	eax
	mov	eax,edx
	and	eax,0x7000
	shr	eax,12

	sub	dword ptr [_bpl1pt],ecx
	cmp	eax,2
	jb	wbplcon0loreskludgeend
	sub	dword ptr [_bpl2pt],ecx
	cmp	eax,3
	jb	wbplcon0loreskludgeend
	sub	dword ptr [_bpl3pt],ecx
	cmp	eax,4
	jb	wbplcon0loreskludgeend
	sub	dword ptr [_bpl4pt],ecx
	cmp	eax,5
	jb	wbplcon0loreskludgeend
	sub	dword ptr [_bpl5pt],ecx
	cmp	eax,6
	jb	wbplcon0loreskludgeend
	sub	dword ptr [_bpl6pt],ecx
	
	
	
wbplcon0loreskludgeend:
	pop	eax
wbplcon0hireskludge:	
wbplcon0nokludge2:	
	pop	ecx        
wbplcon0nokludge:		
#endif
	
	mov     dword ptr [_bplcon0],edx
        shr     edx,12
        and     edx,0fh
        test    byte ptr [_bplcon0 + 1],04h
        jz      nodualpf
        mov     ecx,dword ptr [_decoderoutinedualtab + edx*4]
        jmp     afterdualpf
nodualpf:
        mov     ecx,dword ptr [_decoderoutinetab + edx*4]
afterdualpf:

        mov     dword ptr [_decoderoutineptr],ecx
        mov     dword ptr [_DIWfirstposonline],88
        mov     dword ptr [_DIWlastposonline],472
        test    byte ptr [_bplcon0 + 1],08h
        jz      noham
        mov     ecx,[_drawhamroutine]
        jmp     wakke
noham:  test    byte ptr [_bplcon0 + 1],04h
        jz      nodualpf2
        cmp     edx,1
        je      nodualpf2
        cmp     edx,9
        je      nodualpf2
        mov     ecx,[_drawdualloresroutine]
        cmp     edx,7
        jb      wakke
        mov     ecx,[_drawdualhiresroutine]
        mov     dword ptr _DIWfirstposonline,176
        mov     dword ptr _DIWlastposonline,944
        jmp     wakke
nodualpf2:
        mov     ecx,[_drawloresroutine]
        test    edx,08h
        jz      wakke
        mov     dword ptr _DIWfirstposonline,176
        mov     dword ptr _DIWlastposonline,944
        mov     ecx,[_drawhiresroutine]
wakke:  mov     dword ptr _drawroutineptr,ecx
        call    calculate_window_
        pop     ecx
        ret


;When ddfstrt is (mod 8)+4, shift order is 8-f,0-7, not implemeted in lores yet

;========
; BPLCON1
;========
; $dff102 - WORD

; Extra variables
; oddscroll - dword with the odd lores scrollvalue
; evenscroll -dword with the even lores scrollvalue

; oddhiscroll - dword with the odd hires scrollvalue
; evenhiscroll - dword with the even hires scrollvalue

                ALIGN32

wbplcon1_:
        and     edx,0ffh
        mov     dword ptr _bplcon1,edx
        mov     ecx,edx
        and     edx,0fh
        mov     dword ptr _oddscroll,edx
        and     edx,7h
        test    dword ptr _ddfstrt,2h
        jz      bpc1normal1
        add     edx,4h
        and     edx,7h
bpc1normal1:
        shl     edx,1
        mov     dword ptr _oddhiscroll,edx
        and     ecx,0f0h
        shr     ecx,4
        mov     dword ptr _evenscroll,ecx
        and     ecx,7h
        test    dword ptr _ddfstrt,2h
        jz      bpc1normal2
        add     ecx,4h
        and     ecx,7h
bpc1normal2:

        shl     ecx,1
        mov     dword ptr _evenhiscroll,ecx
        call    calculate_window_
        ret

;========
; BPLCON2
;========
; $dff108 - WORD
                ALIGN32

wbplcon2_:
        mov     dword ptr _bplcon2,edx
        ret
;========
; BPL1MOD
;========
; $dff108 - WORD
                ALIGN32

wbpl1mod_:
        and     edx,0fffeh
        movsx   edx,dx
        mov     dword ptr _bpl1mod,edx
        ret

;========
; BPL2MOD
;========
; $dff10a - WORD
                ALIGN32

wbpl2mod_:
        and     edx,0fffeh
        movsx   edx,dx
        mov     dword ptr _bpl2mod,edx
        ret

; Decide delay of sprite ptr update
; ebx - cycle xpos must be past to trigger delay

                ALIGN32

decidedelay:
        cmp     dword ptr _spritewritereal,1
        je      ddnodelay
        cmp     ebx,dword ptr [_xpos]
        ja      ddnodelay
        mov     ebx,dword ptr [_spritewritenext]
        mov     dword ptr [_spritewritebuffer+ebx],ecx
        mov     dword ptr [_spritewritebuffer+4+ebx],edx
        add     ebx,8
        mov     dword ptr [_spritewritenext],ebx
        mov     ebx,1
        ret
ddnodelay:
        xor     ebx,ebx
        ret

;=======
; SPR0PT
;=======

; $dff120 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff120 - WORD
                ALIGN32

wspr0pth_:
        push    ebx
        mov     ebx,0
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr0h
        pop     ebx
        ret
wspr0h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2,dx
        ret

; $dff122 - WORD
                ALIGN32

wspr0ptl_:
        push    ebx
        mov     ebx,0
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr0l
        pop     ebx
        ret
wspr0l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt,dx
        ret


;=======
; SPR1PT
;=======

; $dff124 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff124 - WORD
                ALIGN32

wspr1pth_:
        push    ebx
        mov     ebx,4
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr1h
        pop     ebx
        ret
wspr1h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+4,dx
        ret

; $dff126 - WORD
                ALIGN32

wspr1ptl_:
        push    ebx
        mov     ebx,4
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr1l
        pop     ebx
        ret
wspr1l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+4,dx
        ret


;=======
; SPR2PT
;=======

; $dff128 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff128 - WORD
                ALIGN32

wspr2pth_:
        push    ebx
        mov     ebx,8
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr2h
        pop     ebx
        ret
wspr2h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+8,dx
        ret

; $dff12a - WORD
                ALIGN32

wspr2ptl_:
        push    ebx
        mov     ebx,8
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr2l
        pop     ebx
        ret
wspr2l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+8,dx
        ret


;=======
; SPR3PT
;=======

; $dff12c - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff12c - WORD
                ALIGN32

wspr3pth_:
        push    ebx
        mov     ebx,12
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr3h
        pop     ebx
        ret
wspr3h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+12,dx
        ret

; $dff12e - WORD
                ALIGN32

wspr3ptl_:
        push    ebx
        mov     ebx,12
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr3l
        pop     ebx
        ret
wspr3l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+12,dx
        ret


;=======
; SPR4PT
;=======

; $dff130 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff130 - WORD
                ALIGN32

wspr4pth_:
        push    ebx
        mov     ebx,16
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr4h
        pop     ebx
        ret
wspr4h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+16,dx
        ret

; $dff132 - WORD
                ALIGN32

wspr4ptl_:
        push    ebx
        mov     ebx,16
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr4l
        pop     ebx
        ret
wspr4l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+16,dx
        ret


;=======
; SPR5PT
;=======

; $dff134 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff134 - WORD
                ALIGN32

wspr5pth_:
        push    ebx
        mov     ebx,20
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr5h
        pop     ebx
        ret
wspr5h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+20,dx
        ret

; $dff136 - WORD
                ALIGN32

wspr5ptl_:
        push    ebx
        mov     ebx,20
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr5l
        pop     ebx
        ret
wspr5l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+20,dx
        ret


;=======
; SPR6PT
;=======

; $dff138 - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff138 - WORD
                ALIGN32

wspr6pth_:
        push    ebx
        mov     ebx,24
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr6h
        pop     ebx
        ret
wspr6h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+24,dx
        ret

; $dff13a - WORD
                ALIGN32

wspr6ptl_:
        push    ebx
        mov     ebx,24
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr6l
        pop     ebx
        ret
wspr6l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+24,dx
        ret


;=======
; SPR7PT
;=======

; $dff13c - konvensjon, peker registre inneholder alltid adresse uten kmem
; $dff13c - WORD
                ALIGN32

wspr7pth_:
        push    ebx
        mov     ebx,28
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr7h
        pop     ebx
        ret
wspr7h: pop     ebx
        and     edx,01fh                ; Store 21 bit (2 MB)
        mov     word ptr _sprpt+2+28,dx
        ret

; $dff13e - WORD
                ALIGN32

wspr7ptl_:
        push    ebx
        mov     ebx,28
        add     ebx,dword ptr [_sprite_delay]
        call    decidedelay
        test    ebx,1
        jz      wspr7l
        pop     ebx
        ret
wspr7l: pop     ebx
        and     edx,0fffeh              ; No odd adresses
        mov     word ptr _sprpt+28,dx
        ret




wdefault_:
        ret




ALIGN32

_TEXT ends

_DATA  segment dword public use32 'DATA'

include data\memory.inc
include data\floppy.inc
include data\sound.inc
include data\inout.inc
include data\joymouse.inc
include data\graphem.inc
include data\bus.inc
include data\blit.inc
include data\draw.inc
	
#ifdef DEBUGBUILD
include data\log.inc
#endif

public _d
public _a
public _aend

_d: dd 8 dup(0)
_a: dd 8 dup(0)
_aend:

_busnoise:       dd      0
                 dd      -1

_mysteryv:	dd   0ff00ff00h
	
_busnoisecounter:        dd      0


extern _k2mem:dword
	
_DATA  ends

_BSS  segment dword public use32 'BSS'
_BSS  ends

end
